;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; Kimmo dictionaries.
;;;

(defclass kimmo-dictionary (filename-symbol-mixin
                            kimmo-string-mixin
                            property-list-mixin)
  ((class-storage  :initform (make-hash-table)
                   :accessor class-storage)
   (lexicon-storage :initform (make-hash-table)
                    :accessor lexicon-storage)
   (lexicon-flavor :initform 'kimmo-lexicon
                   :initarg :lexicon-flavor
                   :accessor lexicon-flavor)
   (root-lexicon-name :initform '|Root|
                      :accessor root-lexicon-name
                      :initarg :root-lexicon-name)
   (root-lexicon-root :accessor root-lexicon-root)
   (comment-delimiters :initform nil
                       :accessor comment-delimiters)
   (lexicon-names :initform nil           ; computed from alternations and used
		  :accessor lexicon-names); for checking actual lexicon names
     )
  ;;(:init-options suppress-initial-read)
  (:default-initargs :allow-other-keys t) ;*** doesn't seem to work? -- sgr
  )

;;; Snort.  Someday pcl::initialize will become initialize-instance! -- sgr
(defmethod pcl::initialize :after ((self kimmo-dictionary) plist)
  (get-defined-value '*kimmo-reader*)
  (unless (getf plist :suppress-initial-read)
    (read-file self)))

(defmethod lexicons ((self kimmo-dictionary))
  ;; return a list of all the lexicons stored in here
  (let ((answer nil))
    (maphash #'(lambda (key value)
                 (declare (ignore key))
                 (push value answer))
             (lexicon-storage self))
    answer))

(defmethod show-lexicon-entries ((self kimmo-dictionary) &optional name)
  ;; print something about each lexicon, or a named lexicon.
  (if name
    (show-entries (retrieve-lexicon self name))
    ;; if no name, do them all
    (loop for lexicon in (lexicons self)
          doing (show-entries lexicon)))
  (values))

(defmethod retrieve-lexicon ((self kimmo-dictionary) name)
  ;; find the lexicon with this name.
  (or (gethash name (lexicon-storage self))
      (error "There is no lexicon named ~S." name)))

(defmethod retrieve-lexicon! ((self kimmo-dictionary) name)
  ;; retrieve the lexicon name name.  If not there but should be, create it.
  (let ((lex (gethash name (lexicon-storage self))))
    (cond ((not (null lex)) lex)
	  ((member name (lexicon-names self))
	   (setq lex (make-lexicon self name))
           (setf (gethash name (lexicon-storage self)) lex)
	   lex)
          (t (error "No lexicon named ~S in the alternations." name)))))

(defmethod make-lexicon ((self kimmo-dictionary) name)
  ;; make a lexicon with some name, that points back to the dictionary
  (make-instance (lexicon-flavor self) :name name :dictionary self))

(defmethod retrieve-class ((self kimmo-dictionary) name)
  ;; find the class with this name
  (gethash name (class-storage self)))

(defmethod retrieve-class-lexicons ((self kimmo-dictionary) name)
  ;; find all lexicons for a given continuation class
  (unless (class-defined self name)
    (error  "Continuation class ~S is undefined." name))
  (mapcar #'(lambda (thing) (retrieve-lexicon self thing))
          (retrieve-class self name)))

(defmethod retrieve-class-lexicon-roots ((self kimmo-dictionary) name)
  ;; find the root of all lexicons of a continuation class
  (mapcar #'root (retrieve-class-lexicons self name)))

(defmethod class-defined ((self kimmo-dictionary) name)
  ;; is there a class with this name?
  (multiple-value-bind (value found-p)
                       (gethash name (class-storage self))
    (declare (ignore value))
    found-p))

;;;
;;; File Input
;;;

(defmethod process-file-alternations ((self kimmo-dictionary) stream)
  ;; first, clear state out of the reader and self
  (disable-digits *kimmo-reader*)
  (disable-special-return *kimmo-reader*)
  (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
         '( #\: #\;)) ; #\.
  (clrhash (class-storage self))
  (setf (lexicon-names self) nil)
  (setf (comment-delimiters self) nil)
  (setf (accent-codes self) nil)
  (disable-comments *kimmo-reader*)
  (let ((token (read-next-thing *kimmo-reader* stream nil))) ;error if EOF
    ;; This parsing code isn't very general.  The order %COMMENT-DELIMITERS,
    ;; %ACCENTS, ALTERNATIONS is required.  Also, the %ACCENTS declaration
    ;; must be all on one line.
    (when (eq token '%COMMENT-DELIMITERS)
      ;; the file tells us what kind of comments to look out for.
      (let ((open  (coerce (read-next-thing *kimmo-reader* stream nil)
                           'character))
            (close (coerce (read-next-thing *kimmo-reader* stream nil)
                           'character)))
        (when (char= open close)
          (file-syntax-error self "Open and close %COMMENT-DELIMITERS identical."))
        (setf (comment-delimiters self) (list open close))
        (enable-comments *kimmo-reader* open close)
        (setq token (read-next-thing *kimmo-reader* stream nil))))
    (when (eq token '%ACCENTS)
      ;; the language supports these accents.
      (enable-special-return *kimmo-reader* '*CR*)
      (let ((accents (read-until *kimmo-reader* '*CR* stream nil)))
        (disable-special-return *kimmo-reader*)
        (setf (accent-codes self) 
              (loop for x in accents 
                    collecting (aref (string x) 0)))) ;(exploden x))))
      (setq token (read-next-thing *kimmo-reader* stream nil)))
    (unless (eq token 'ALTERNATIONS)
      ;; the rest of it must be a set of alternations.
      (file-syntax-error self "File doesn't begin with ALTERNATIONS declaration."))
    (loop with class and def and equal-sign
          for alt = (read-next-thing *kimmo-reader* stream nil)
          until (eq alt 'END)
          do (unless (and (listp alt) (eq (cadr alt) '=))
               (file-syntax-error self "~S is an invalid class definition." alt))
             (setq class      (first alt)
                   equal-sign (second alt)
                   def        (cddr alt))
             (unless (eq equal-sign '=)
               (file-syntax-error self "Bad class definition for ~S." class))
             (when (gethash class (class-storage self))
               (file-syntax-error self "Duplicate class definition for ~S." class))
             ;; make this class-name point to these things.
             (setf (gethash class (class-storage self)) def))
    (do-alternation-closure self)))

(defmethod do-alternation-closure ((self kimmo-dictionary))
  (let ((aux (make-hash-table))
	;;(u (make-mcflavor-instance 'simple-set-universe)) ;***
	(classes nil)
	(mentioned nil)
	(directly-recursive nil)
	lexicon-names* ;star denotes transitive closure...
	lexicon-names-set)
    (maphash #'(lambda (key value)
                 ;; aux is initially a copy of class-storage.
                 (setf (gethash key aux) value);;(encode u value))
                 ;; make a list of the classes and which ones are mentioned
                 (push key classes)
                 (push key mentioned)
                 (when (member key value)
                   ;; note classes that directly include themselves.
                   (pushnew key directly-recursive)))
             (class-storage self))
    (loop for changed = nil
          ;; keep looping as long as something changes
       do (loop for c in classes
	     as rhs = (gethash c aux) ;defn of class c
	     as xrhs = rhs ;;(decode u rhs)
	     as nrhs = nil
	     do (loop for y in xrhs
		   as ydef = (gethash y aux) ;defn of y, member of rhs of x
		   do (pushnew y mentioned)
                      ;; add the definition of y to that of x (i.e., we're 
                      ;; taking the transitive closure of continuation classes)
                      (if (null ydef)
                        ;; this doesn't have a right-hand side
                        ;; (setq nrhs (add-one u y nrhs)
                        (pushnew y nrhs)
                        ;; this does have a right-hand side
                        ;;(setq nrhs (union u nrhs ydef))
                        (setq nrhs (union nrhs ydef :test #'equal))))
             unless (not (set-exclusive-or nrhs rhs :test #'equal))
               ;; xor gives set of elts in exactly one; null means they're same!
               ;; the rhs of x has changed, so we have to go 'round again
               do (setq changed t)
                  (setf (gethash c aux) nrhs))
       while changed)
    (loop for m in mentioned
       when (or (member m directly-recursive)
                (not (gethash m aux)))
         do (push m lexicon-names*))
    (setf (lexicon-names self) lexicon-names*)
    (setq lexicon-names-set lexicon-names*);;(encode u lexicon-names*))
    (maphash #'(lambda (key value)
                 (setf (gethash key (class-storage self))
                       (intersection value lexicon-names-set :test #'equal)
                       ;(decode u
                       ;        (intersection u 
                       ;                      value lexicon-names-set))
                       ))
             aux)))

(defmethod process-file-lexicon-entries ((self kimmo-dictionary) stream)
  (clrhash (lexicon-storage self))
  (let ((token (read-next-thing *kimmo-reader* stream nil)))
    (unless (eq token 'LEXICON)
      (file-syntax-error self "Lexicon section doesn't start with LEXICON."))
    (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
          '( #\: #\;)) ;#\.
    (loop with lexicon
          do (ecase token
               (LEXICON
                (setq lexicon (retrieve-lexicon!
                               self 
                               (read-next-thing *kimmo-reader* stream nil)))
                (process-file-entry lexicon stream))
               (\;
                (process-file-entry lexicon stream))
               (END 
                (return nil))
               (%GENERATOR-DEMO
                (file-syntax-error self 
                                   "GENERATOR-DEMO belongs in the .AUT file."))
               (%RECOGNIZER-DEMO
                ;; give it a certain %recognizer-demo property.
                (setf (getprop self token)
                      (read-until *kimmo-reader* 'END stream nil))))
          (setq token (read-next-thing *kimmo-reader* stream nil)))))

(defmethod zap-continuation-classes ((self kimmo-dictionary))
  (mapc #'zap-continuation-classes (lexicons self)))

(defmethod read-file ((self kimmo-dictionary))
  (with-open-stream (stream (input-stream self))
    (format t "~&Reading ~A...~%" (filename self))
    (process-file-alternations self stream)
    (process-file-lexicon-entries self stream)
    (zap-continuation-classes self)
    (setf (root-lexicon-root self)
	  (root (retrieve-lexicon self (root-lexicon-name self))))
    (enable-special-return *kimmo-reader* '*CR*) ;; cleanup
    self))
