;;; -*- Mode: Lisp; Package: (KIMMO :use Lisp); Syntax: Common-Lisp -*-
;;; Created January 1989 by sgr.

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice

;;; The Kimmo Morphological Analyzer.  To port this, search forward for *** and
;;; change the value of the variable *KIMMO-DIRECTORY*.  The rest is just like
;;; PCL; that is, look at kimmo::compile-kimmo and kimmo::load-kimmo.
;;;
;;; Also, search forward for *** to find *default-language-directory*, and 
;;; change that, too.
;;;

#-Symbolics
(require "LOOP" "/mit/6.863/loop/loop") ;***

(require "PCL"  "/mit/6.863/pcl/pcl")   ;***

;;;
;;; Now define the Kimmo package.  The reason this looks complicated is to
;;; preserve the behavior of LOOP in pure CL packages, i.e., we have to shadow
;;; the LOOP we inherit from the LISP package.
;;;

(in-package "KIMMO" :use (list (find-package "PCL")
                               (find-package "LISP")))
#-Symbolics
(shadowing-import 'loop:loop "KIMMO")
#-Symbolics
(use-package (find-package "LOOP") "KIMMO")

;;;
;;; System definition for Kimmo.  This uses the "portable defsystem" that 
;;; comes with PCL.  (Thank you, Gregor!)  See the PCL source for documentation.
;;;

;;; "Kimmo;" logical pathname defined in the ccl;library: file.
#-Symbolics
(defvar *kimmo-directory* "/mit/6.863/kimmo/") ;*** Change this when porting.

#-Symbolics
(defmacro def-simple-system (name directory &body files)
  ;; define a simple system with linear dependencies, i.e., "compile and load
  ;; the files in this order"
  `(pcl::defsystem ,name ,directory
     ,(let ((answer nil)
            (files-already-seen nil))
        (dolist (file files)
          (let ((files-file-depends-on (copy-list files-already-seen)))
            ;; make this file depend (in all ways) on everything seen up to now
            (push `(,file ,files-file-depends-on
                          ,files-file-depends-on
                          ,files-file-depends-on)
                  answer))
          (push file files-already-seen))
        (nreverse answer))))

#-Symbolics
(def-simple-system kimmo *kimmo-directory*
  ;; the file dependencies of Kimmo.
  "Kimmo-Defsystem"
  "compat" "reader" "letter-tree"  
  "filename-symbol-mixin" "string"
  "lexicon" "dictionary" "merged-dictionary"
  "automaton" "automaton-set"
  "machine" "generator" "recognizer" 
  ;; "test-kimmo" ;test suite yet to be written
  )

#+Symbolics
(sct:defsystem kimmo
    (:default-pathname "NLP:Kimmo;")
  (:serial "Kimmo-Defsystem"
           "compat" "reader" "letter-tree" "filename-symbol-mixin" "string"
	   "lexicon" "dictionary" "merged-dictionary"
	   "automaton" "automaton-set"
	   "machine" "generator" "recognizer"))

;;;
;;; Tools for loading and compiling Kimmo.  These are in imitatione of PCL.
;;;

#-Symbolics
(defun load-kimmo ()
  ;; load the kimmo system
  (pcl::operate-on-system 'kimmo :load)
  (values))

#-Symbolics
(defun compile-kimmo ()
  ;; compile the kimmo system
  (pcl::operate-on-system 'kimmo :compile)
  (values))

#-Symbolics
(defun edit-kimmo ()
  ;; get all the kimmo files in editor buffers.
  (flet ((system-modules (system-name)
           (cadr (pcl::get-system system-name)))
         (module-source-file (module directory)
           (let ((pcl::*system-directory* directory))
             (pcl::make-source-pathname (pcl::module-name module)))))
    (dolist (module (system-modules 'kimmo))
      (ed (module-source-file module *kimmo-directory*)))
    (values)))

;;;
;;; Tools for reading in language definitions.
;;;

(defparameter *default-language-directory* "/mit/6.863/kimmo/languages/") ;*** change in porting
						;
(defvar *languages* nil "Languages that have been read.")

(defun read-language-definition (language &optional (directory *default-language-directory*))
  ;; read in a particular language
  (declare (values dictionary-instance automaton-set-intance))
  (let ((dictionary-file  (probe-file (merge-pathnames
					(format nil "~(~A~).dic" language) 
					directory))))
    (when (null dictionary-file)
      ;; can't find the dictionary, so punt early.
      (format *debug-io* "~&There is no .DIC file for ~:(~A~)." language)
      (return-from read-language-definition (values)))
    (let* ((automata-file (probe-file (merge-pathnames
				       (format nil "~(~A~).aut" language)
				       directory))))
      (when (null automata-file)
	;; can't find the automata file, so punt early.
	(format *debug-io* "~&There is no .AUT file for ~:(~A~)." language)
	(return-from read-language-definition (values)))
      ;; just one automata file, so read dictionary & automata
      (let ((dictionary (make-instance 'kimmo-dictionary 
                                       :filename dictionary-file))
            (automata   (make-instance 'kimmo-automaton-set 
                                       :filename automata-file
                                       :dictname (intern 
                                                  (format
                                                   nil
                                                   "~:@(~A~)-DIC"
                                                   (pathname-name dictionary-file)
                                                   "DIC")))))
        (pushnew (string-upcase (string language)) *languages*)
        (return-from read-language-definition
                     (values dictionary automata))))))

(defun read-all-language-definitions
       (&optional (directory *default-language-directory*))
  ;; read in the dictionaries and automata in directory
  (loop with dict and aut
        for dictionary-file in (directory (merge-pathnames directory "*.dic"))
        doing (multiple-value-setq (dict aut)
                (read-language-definition (pathname-name dictionary-file) directory))
        when (and dict aut)
          collect dict into dictionaries
          and collect aut into automata
        finally (return (values dictionaries automata))))

(provide 'kimmo)
