;;; -*- Mode: Lisp; Package: Kimmo; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package 'kimmo)

;;;
;;; Filtered letter trees.
;;;

(defclass filtered-letter-tree (letter-tree) ())

(defstruct (filtered-letter-tree-node 
            (:type list) 
            (:include letter-tree-node))
  ;; The interpretation of the filter bits is that the filter bits
  ;; supplied in a call must have a non-null INTERSECTION with the
  ;; filter bits stored in the tree.
  (filter-bits #* )) ;0-extension means this is all zeros, effectively

(defmethod node-outputs ((self filtered-letter-tree) node &optional filter-bits)
  (loop for out in (letter-tree-node-outputs node)
        when (bitvectors-overlap-p (car out) filter-bits)
          collect (cdr out)))

(defmethod raw-node-outputs ((self filtered-letter-tree) node filter-bits)
  (loop for out in (letter-tree-node-outputs node)
        when (bitvectors-overlap-p (car out) filter-bits)
          collect out))

(defmethod all-node-outputs ((self filtered-letter-tree) node)
  (mapcar #'cdr (letter-tree-node-outputs node)))

(defmethod all-raw-node-outputs ((self filtered-letter-tree) node)
  ;; returns pairs (bits . output)
  (letter-tree-node-outputs node))

(defmethod possible-branches ((self filtered-letter-tree) node)
  (node-branches self node))

(defmethod filtered-branches ((self filtered-letter-tree) node filter-bits)
  (loop for branch in (letter-tree-node-subtrees node)
        when (bitvectors-overlap-p 
              (filtered-letter-tree-node-filter-bits branch)
              filter-bits)
       collect branch))

(defmethod filter-branch ((self filtered-letter-tree) branch filter-bits)
  (when (bitvectors-overlap-p (filtered-letter-tree-node-filter-bits branch)
                              filter-bits)
    branch))

(defmethod get-branch ((self filtered-letter-tree) char node &optional filter-bits)
  (let ((branch (assoc char (letter-tree-node-outputs node))))
    (when (and branch
               (bitvectors-overlap-p 
                (filtered-letter-tree-node-filter-bits branch)
                filter-bits))
      branch)))

(defmethod pcl::initialize :after ((self filtered-letter-tree) plist)
  (declare (ignore plist))
  (reset self))

(defmethod make-node ((self filtered-letter-tree) key)
  (make-filtered-letter-tree-node :key key))

(defmethod add-output ((self filtered-letter-tree) char-list output &optional filter-bits)
  (let* ((node (add-char-list self char-list filter-bits))
	 (outlist (letter-tree-node-outputs node)))
    (loop for out in outlist
          when (eq (cdr out) output)
            do (setf (car out) (bitvector-or (car out) filter-bits))
               (return output)
         finally (setf (letter-tree-node-outputs node)
                       (nconc outlist (list (cons filter-bits output))))
                 (return output))))

(defmethod look-up-char-list ((self filtered-letter-tree) char-list 
                              &optional root-subtree filter-bits) ;arg change
  (check-type filter-bits '(vector bit))
  (loop with tree = (or root-subtree (root self))
        for x in char-list
        as subtrees = (letter-tree-node-subtrees tree)
        as subtree = (assoc x subtrees)
        do (cond ((eql x (null-char self))
                  ;; not a true character; ignore
                  nil)
                 ((null subtree)
                  (return nil))
                 ((not (bitvectors-overlap-p 
                        (filtered-letter-tree-node-filter-bits subtree)
                        filter-bits))
                  (return nil))
                 (t (setq tree subtree)))
        finally (return tree)))

(defmethod add-char-list ((self filtered-letter-tree) char-list 
                          &optional root-subtree filter-bits)
  ;; no &optional root-subtree here; harder with the filter bits
  (declare (ignore root-subtree))
  (loop with tree = (root self)
        for x in char-list
        as subtrees = (letter-tree-node-subtrees tree)
        as subtree = (assoc x subtrees)
        do (cond ((eql x (null-char self))
                  ;; not a true character; ignore
                  nil)
                 ((null subtree)
                  (setq subtree (make-node self x))
                  (splice-in-subtree self tree subtree)
                  (setq tree subtree))
                 (t (setq tree subtree)))
           (setf (filtered-letter-tree-node-filter-bits tree)
                 (bitvector-or (filtered-letter-tree-node-filter-bits tree)
                               filter-bits))
        finally (return tree)))

(defmethod print-outputs ((self filtered-letter-tree) &rest keywords
                          &key filter-bits
                          &allow-other-keys)
  (let ((filter-seg 
         (and filter-bits
              (list :outputs-function 
                    #'(lambda (x) (node-outputs self x filter-bits))
                    :subtrees-function
                    #'(lambda (x) (filtered-branches self x filter-bits))))))
      (apply #'map-output-lists
             self
             #'(lambda (char-list outputs)
                 (format t 
                         (format-template 
                          "~&" (:use-repeatedly-on-list "~A")
                          ":~20T" (:separated-list "~S" ", ")
                          "~%")
                         char-list
                         outputs))
             :supply-char-lists t
             (append filter-seg keywords))))

;;;
;;; Merged dictionaries.
;;;

(defclass merged-kimmo-dictionary (kimmo-dictionary)
  ((lexicon-flavor :initform 'virtual-kimmo-lexicon
                   :accessor lexicon-flavor
                   :initarg :lexicon-flavor)
   (letter-tree-flavor :initform 'filtered-letter-tree
                       :accessor letter-tree-flavor
                       :initarg :letter-tree-flavor)
   (merged-letter-tree :accessor merged-letter-tree
                       :initarg :merged-letter-tree)
   (lexicon-universe :initform (make-instance 'simple-set-universe) ;***
                     :accessor lexicon-universe
                     :initarg :lexicon-universe)))

(defmethod pcl::initialize :after ((self merged-kimmo-dictionary) plist)
  (declare (ignore plist))
  (setf (merged-letter-tree self) (make-instance (letter-tree-flavor self))))

(defmethod read-file :before ((self merged-kimmo-dictionary))
  (reset (merged-letter-tree self))
  (reset (lexicon-universe self))) ;***

(defmethod encode-lexicon-set ((self merged-kimmo-dictionary) x)
  (encode (lexicon-universe self) x)) ;***

(defmethod decode-lexicon-set ((self merged-kimmo-dictionary) x)
  (decode (lexicon-universe self) x)) ;***

(defmethod format-lexicon-set ((self merged-kimmo-dictionary) 
                               x &optional (stream *standard-output*))
  (princ #\{ stream)
  (print-separated-list (decode (lexicon-universe self) x) t ", " stream) ;***
  (princ #\} stream))

(defmethod make-lexicon ((self merged-kimmo-dictionary) name)
  (make-virtual-lexicon self (list name `(:lexicon ,name))))

(defmethod retrieve-class-lexicons ((self merged-kimmo-dictionary) name)
  ;; The class storage will now map to virtual lexicons.
  (unless (class-defined self name)
    (error "Continuation class ~S is undefined." name))
  (let ((lexicon (retrieve-class self name)))
    ;; special case:  don't return any lexicons at all for #
    (if (null (lexicon-names lexicon))
      nil
      (list lexicon))))

(defmethod show-entries ((self merged-kimmo-dictionary)
                         &optional (filter-bits -1)) ;***
  ;; An output is initially a list (continuation-class-name features string),
  ;; but the class name is later replaced by a list (class-name . lexicon-roots)
  (unless (bitset-complemented filter-bits) ;***
    (format t "~&Entries for ")
    (format-lexicon-set self filter-bits)
    (format t ":~%"))
  (map-output-lists (slot merged-letter-tree)
                    #'show-raw-output-list
                    :subtrees-function 
                    #'(lambda (x) (filtered-branches self x filter-bits))
                    :outputs-function
                    #'(lambda (x) (raw-node-outputs self x filter-bits))))

(defmethod show-raw-output-list ((self merged-kimmo-dictionary) outputs)
  (loop for (bits cclass feats string) in outputs
     do (format t "~&~A" string)
        (format t "  ~15T~A  ~25T~:A  ~60T"
                (if (listp cclass) ;already expanded
                  (car cclass)
                  cclass)
                (reverse feats))
	(format-lexicon-set self bits t)
	(format t "~%")))

(defmethod zap-continuation-classes ((self merged-kimmo-dictionary))
  (map-output-lists (merged-letter-tree self)
                    #'zap-output-list-continuation-classes))

(defmethod zap-output-list-continuation-classes ((self merged-kimmo-dictionary)
                                                 outputs)
  (loop for out in outputs
     unless (listp (car out))	; already expanded
       do (setf (car out) 
                (cons (car out)
                      (retrieve-class-lexicon-roots self (car out))))))

(defmethod make-virtual-lexicon ((self merged-kimmo-dictionary) def description)
  (let ((bits (encode (lexicon-universe self) def))) ;***
    (make-instance (lexicon-flavor self)
	    :lexicon-names def
	    :lexicon-bits bits
	    :merged-letter-tree (merged-letter-tree self)
            :dictionary self
	    :description description)))

(defmethod do-alternation-closure :after ((self merged-kimmo-dictionary))
  ;; After the alternation closure sets the class storage to map
  ;; every class name to a list of lexicon names, we'll go back
  ;; and replace the list by a virtual lexicon.
  (maphash #'(lambda (class-name def)
               ;; cf CLtL, p. 285.
               (setf (gethash class-name (class-storage self))
                     (make-virtual-lexicon self def `(:class ,class-name))))
           (class-storage self)))

;;;
;;; Virtual lexicons.
;;;

(defclass virtual-kimmo-lexicon (kimmo-lexicon-base-mixin)
  ((lexicon-names :initarg :lexicon-names
                  :accessor lexicon-names)
   (lexicon-bits :initarg :lexicon-bits
                 :accessor lexicon-bits)
   (merged-letter-tree :initarg :merged-letter-tree
                       :accessor merged-letter-tree)
   (root :initarg :root
         :accessor root)
   (description :initarg :description
                :accessor description)))

(defmethod pcl::initialize :before ((self virtual-kimmo-lexicon) plist)
  ;; tagged private version of root
  (declare (ignore plist))
  (setf (root self)
	(cons self (cdr (root (merged-letter-tree self)))))) ;*** bound now?

(defmethod pcl::initialize :after ((self virtual-kimmo-lexicon) plist)
  (declare (ignore plist))
  (let ((d (or (description self)
               (error "Description should be non-NIL?"))))
    (when (eq d ':random)
      (setf (description self)
	    (setq d '(:random))))
    (ecase (car d)
      (:lexicon (setf (name self) (cadr d)))
      (:class   (setf (name self) (format nil "(~A)" (cadr d))))
      (:random  (setf (name self) (format-lexicon-set 
                                   self (lexicon-bits self) nil))))))

(defmethod reset ((self virtual-kimmo-lexicon))
  (error "Can't reset a virtual lexicon."))

(defmethod map-output-lists ((self virtual-kimmo-lexicon) function &rest kwds)
  (apply #'map-output-lists
         (merged-letter-tree self)
         function
         kwds))

(defmethod node-outputs ((self virtual-kimmo-lexicon) node)
  (node-outputs (merged-letter-tree self) node (lexicon-bits self)))

(defmethod possible-branches ((self virtual-kimmo-lexicon) node)
  (possible-branches (merged-letter-tree self) node))

(defmethod filtered-branches ((self virtual-kimmo-lexicon) node)
  (filtered-branches (merged-letter-tree self) node (lexicon-bits self)))

(defmethod filter-branch ((self virtual-kimmo-lexicon) branch)
  (filter-branch (merged-letter-tree self) branch (lexicon-bits self)))

(defmethod get-branch ((self virtual-kimmo-lexicon) char node)
  (get-branch (merged-letter-tree self) char node (lexicon-bits self)))

(defmethod show-entries ((self virtual-kimmo-lexicon))
  (show-entries (dictionary self) (lexicon-bits self)))

(defmethod add-output ((self virtual-kimmo-lexicon) char-list output)
  (unless (eq (car (description self)) ':lexicon)
    (error "Can't add outputs to non-lexicon class ~S." self))
  (add-output (merged-letter-tree self) char-list output (lexicon-bits self)))

;; File Input just uses add-output method

