;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; Object somewhat like a readtable.  This makes it possible to do some things
;;; CL doesn't permit with real readtables, viz. reading lowercase symbols, and
;;; turning off integers.  Cf. CLtL, pp. 361-362.
;;;

(defparameter *digits*      '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
(defparameter *whitespaces* '(#\Tab #\Space #\Page #\Newline #\Return #\Linefeed))
(defparameter *delimiters*  '(#\;))
(defparameter *eof-token*   (make-symbol "EOF"))

(defclass kimmo-reader (print-object-mixin)
  ((digit-chars :initform (copy-list *digits*)
                :initarg :digit-chars
                :accessor digit-chars)
   (whitespace-chars :initform (copy-list *whitespaces*)
                     :initarg :whitespace-chars
                     :accessor whitespace-chars)
   (delimiter-chars :initform (copy-list *delimiters*)
                    :initarg :delimiter-chars
                    :accessor delimiter-chars)
   (open-comment :initform nil
                 :initarg :open-comment
                 :accessor open-comment)
   (close-comment :initform nil
                  :initarg :close-comment
                  :accessor close-comment)
   ;; if non-NIL, this is what #\Return yields
   (return-token :initform nil
                 :initarg :return-token
                 :accessor return-token)
   ;; these slots never change.  see be-paranoid-about, below.
   (open-paren :initform #\(
               :initarg :open-paren
               :reader open-paren)
   (close-paren :initform #\)
                :initarg :close-paren
                :reader close-paren)
;;; This doesn't use dotted lists, but *DOES* want #\. to be alphabetic!
;  (cons-dot :initform #\.
;            :initarg :cons-dot
;            :reader cons-dot)
   (escape :initform #\\ 
           :initarg :escape
           :accessor escape) ;ok, so we *DO* change this one...
   (string-quote :initform #\"
                 :initarg :string-quote
                 :reader string-quote)
   (name :initform "Un-named"
         :initarg :name
         :reader name)
   (package :initform *package*
            :initarg :package
            :reader package)))

(defmethod print-object-name ((self kimmo-reader))
  ;; the idiosyncratic part of this thing for printing
  (name self))

(defmethod be-paranoid-about ((self kimmo-reader) char type)
  ;; see if we think it's reasonable for char to be of type -- type is one
  ;; of digit, whitespace, delimiter, open/close comment.  Can't allow those
  ;; to collide with the other slots whose value we never change.
  (check-type char character)
  (cond ((or (and (characterp (open-paren self))
                  (char= char (open-paren self)))
             (and (characterp (close-paren self))
                  (char= char (close-paren self))))
         (error "~S can't be ~A because it's a paren: ~S or ~S."
                char type (open-paren self) (close-paren self)))
        ;; It turns out that there *ARE* times we want to make . alphabetic... ???
;        ((char= char (cons-dot self))
;         (error "~S can't be ~A because it's a cons-dot: ~S."
;                char type (cons-dot self)))
        ((and (characterp (escape self))
              (char= char (escape self)))
         (error "~S can't be ~A because it's an escape char: ~S."
                char type (escape self)))
        ((char= char (string-quote self))
         (error "~S can't be ~A because it's a string-quote: ~S."
                char type (string-quote self)))))

;;; *** This shouldn't be called, except for single-char comment openers?
;;;     No element of a comment opener/closer can be a digit, delim, etc.?
(defmethod be-paranoid-about-comments ((self kimmo-reader) char type)
  ;; additional paranoia to be enforced for comment openers & closers
  (be-paranoid-about self char type)
  (cond ((member char (digit-chars self) :test #'char=)
         (error "~S can't be ~A because it's a digit." char type))
        ((member char (whitespace-chars self) :test #'char=)
         (error "~S can't be ~A because it's a whitespace." char type))
        ((member char (delimiter-chars self) :test #'char=)
         (error "~S can't be ~A because it's a delimiter." char type))))

;;;
;;; Managing digits.
;;;

(defmethod enable-digit-char ((self kimmo-reader) char)
  ;; add char to the set of digits
  (be-paranoid-about self char "a digit")
  (pushnew char (digit-chars self)))

(defmethod disable-digit-char ((self kimmo-reader) char)
  ;; make char no longer a digit
  (setf (digit-chars self) (delete char (digit-chars self) :test #'char=)))

(defmethod enable-digits ((self kimmo-reader) &optional (digits *digits*))
  ;; restore the some digits
  (loop for char in digits
        doing (be-paranoid-about self char "a digit"))
  (setf (digit-chars self) digits))

(defmethod disable-digits ((self kimmo-reader))
  ;; make all digits go away
  (setf (digit-chars self) nil))

(defmethod char-is-digit-p ((self kimmo-reader) char)
  ;; check if this character is a digit
  (member char (digit-chars self) :test #'char=))

(defmethod radix ((self kimmo-reader))
  ;; the radix is the number of digits!
  (length (digit-chars self)))

;;;
;;; Managing whitespace.
;;;

(defmethod enable-whitespace-char ((self kimmo-reader) char)
  ;; add char to the set of whitespace chars
  (be-paranoid-about self char "a whitespace")
  (when (char= char #\Return)
    ;; if this is a return, then we want to clear this slot, too
    (setf (return-token self) nil))
  (pushnew char (whitespace-chars self)))

(defmethod disable-whitespace-char ((self kimmo-reader) char)
  ;; make char no longer whitespace
  (setf (whitespace-chars self) (delete char (whitespace-chars self) :test #'char=)))

(defmethod enable-whitespaces ((self kimmo-reader) &optional (whitespaces *whitespaces*))
  ;; restore the some whitespace chars
  (loop for char in whitespaces
        doing (be-paranoid-about self char "a whitespace"))
  (setf (whitespace-chars self) whitespaces))

(defmethod disable-whitespaces ((self kimmo-reader))
  ;; make all the whitespaces go away
  (setf (whitespace-chars self) nil))

;;;
;;; Managing delimiters.
;;;

(defmethod enable-delimiter-char ((self kimmo-reader) char)
  ;; add char to the set of delimiter chars
  (be-paranoid-about self char "a delimiter")
  (pushnew char (delimiter-chars self)))

(defmethod disable-delimiter-char ((self kimmo-reader) char)
  ;; make char no longer a delimiter
  (setf (delimiter-chars self) (delete char (delimiter-chars self) :test #'char=)))

(defmethod enable-delimiters ((self kimmo-reader) &optional (delimiters *delimiters*))
  ;; restore some delimiter chars
  (loop for char in delimiters
        doing (be-paranoid-about self char "a delimiter"))
  (setf (delimiter-chars self) delimiters))

(defmethod disable-delimiters ((self kimmo-reader))
  ;; make all the delimiters go away
  (setf (delimiter-chars self) nil))

;;;
;;; Managing comments.
;;;

;;; ***
(defmethod enable-comments ((self kimmo-reader) open-char close-char)
  ;; make these two characters the open and close comment delimiters
  (when (char= open-char close-char)
    (error "Open and close comment delimiters are the same: ~S." open-char))
  ;; dump existing comment delimiters
  (disable-comments self) 
  (be-paranoid-about-comments self open-char "a comment opener")
  (be-paranoid-about-comments self close-char "a comment closer")
  ;; install them
  (setf (open-comment self) open-char
        (close-comment self) close-char))

(defmethod disable-comments ((self kimmo-reader))
  ;; make comments impossible
  (setf (open-comment self)  nil
        (close-comment self) nil))

;;;
;;; Managing alphabetic chars.
;;;

(defmethod enable-alphabetic-char ((self kimmo-reader) char)
  ;; make sure this char is alphabetic
  (be-paranoid-about self char "an alphabetic")
  (when (or (and (characterp (open-comment self))
                 (char= char (open-comment self))) ;***
            (and (characterp (close-comment self))
                 (char= char (close-comment self))))
    (error "~S can't be alphabetic, because it opens or closes a comment: ~S, ~S."
           (open-comment self) (close-comment self)))
  ;; alphabetics are things that aren't anything else.
  (disable-digit-char self char)
  (disable-whitespace-char self char)
  (disable-delimiter-char self char))

;;;
;;; Making return special, turning off backslashes, etc.
;;;

(defmethod enable-special-return ((self kimmo-reader) &optional (token '*cr*))
  ;; make returns read as this
  (enable-alphabetic-char self #\Return)
  (setf (return-token self) token))

(defmethod disable-special-return ((self kimmo-reader))
  ;; make returns be whitespace again
  (enable-whitespace-char self #\Return))

(defmethod disable-escapes ((self kimmo-reader))
  ;; turn off \'s
  (setf (escape self) nil))

(defmethod enable-escapes ((self kimmo-reader))
  ;; turn \'s back on
  (setf (escape self) #\\))

;;;
;;; How to do the actual reading.
;;;

(defun peek-char-really (stream eof-value)
  ;; get around a Lucid bug in peek-char -- it goes into the debugger 
  ;; even when eof-error-p is nil.
  (if (listen stream)
      (peek-char nil stream (null eof-value) eof-value t)
    eof-value))

(defmethod read-next-thing ((self kimmo-reader)
                            &optional
                            (stream *standard-input*)
                            (eof-value *eof-token*))
  ;; returns next thing, and its type (:list, #||:cons-dot||#, :string, 
  ;; :delimiter, :symbol, :return-token or :close-comment)
  (declare (values next-thing next-thing-type))
  (loop while (listen stream)
	for next-char = (peek-char-really stream eof-value)
        ;; main dispatch for the reader.
        if (char= next-char (open-paren self))
          ;; read a list.  Can't use read-delimited-list, since that
          ;; calls read.  Want a version of read-delimited list to which
          ;; we can pass our own reader function???
          return (read-next-list self stream eof-value)
        else if (and (char= next-char #\Newline)
                     (return-token self))
          do (read-char stream (null eof-value) eof-value t)
          and return (values (return-token self)
                             :return-token)
        else if (char= next-char (close-paren self))
          ;; end of a list
          return (values (read-char stream (null eof-value) eof-value t)
                         :close-paren)
        else if (and (characterp (open-comment self)) ;***
                     (char= next-char (open-comment self)))
          ;; flush a comment.
          do (flush-next-comment self stream eof-value)
        else if (and (characterp (close-comment self))
                     (char= next-char (close-comment self))) ;***
          ;; end of a comment (this should never be used, actually...)
          return (values (read-char stream (null eof-value) eof-value t)
                         :close-comment)
;;; Turns out this doesn't make dotted lists, but *DOES* want . to be a delim
;       else if (char= next-char (cons-dot self))
;         ;; unlike the lisp reader, .foo is malformed.  
;         ;; escape the . if you don't like it.
;         return (values (read-char stream (null eof-value) eof-value t)
;                        :cons-dot)
        else if (char= next-char (string-quote self))
          ;; read a string
          return (read-next-string self stream eof-value)
        else if (member next-char (delimiter-chars self) :test #'char=)
          ;; found a delimiter
          return (values (read-char stream (null eof-value) eof-value t)
                         :delimiter)
        else if (and (characterp (escape self))
                     (char= next-char (escape self)))
          ;; found an escape character.  flush the next char & read a symbol
          do (read-char stream (null eof-value) eof-value t)
          and return (read-next-symbol self stream eof-value)
        else if (member next-char (whitespace-chars self) :test #'char=)
          ;; flush whitespace to prevent it from getting into symbols.
          do (read-char stream (null eof-value) eof-value t)
        else 
          ;; found a random char, so accumulate a symbol
          return (read-next-symbol self stream eof-value)))

(defmethod read-next-list ((self kimmo-reader) stream eof-value)
  ;; readtable-object's version of read-delimited-list.
  (loop with dotted-tail = nil
        with past-dotted-tail = nil
        ;; flush the open-paren
        initially (read-char stream (null eof-value) eof-value t)
        ;; ignore second value, only interestedin consing dots & close-parens.
        for next-item = (read-next-thing self stream eof-value)
        if (eql next-item (close-paren self))
          ;; found the close-paren, so return the list
          return (values (if dotted-tail
                           ;; force the tail to be dotted.
                           (apply #'list* items)
                           items)
                         :list)
;;; It turns out this doesn't make dotted lists, but it *DOES* use . as a delim
;        else if (eql next-item (cons-dot self))
;          ;; found dot, so do elementary error check
;          ;; I give up trying to get loop if-then-else to nest right!
;          do (when dotted-tail
;               (error "More than one cons dot?"))
;             (when (null items)
;               (error "Cons dot before first list element?"))
;             (setq dotted-tail t)
        else 
          ;; otherwise just collect it
          if past-dotted-tail
            do (error "More than one item after cons dot?")
          else 
            collect next-item into items
            and when dotted-tail
                 ;; only 1 item after a consing dot.
                 do (setq past-dotted-tail t)))

(defmethod read-next-string ((self kimmo-reader) stream eof-value)
  ;; readtable-object's version of read-delimited-string
  (read-char stream (null eof-value) eof-value t) ;flush the open-quote
  (loop for char = (read-char stream (null eof-value) eof-value t)
        if (and (characterp (escape self))
                (char= char (escape self)))
          ;; found an escape char, so the *NEXT* char definitely goes in
          collect (read-char stream (null eof-value) eof-value t) into string
        else if (char= char (string-quote self))
          ;; found the end of the string (and it's not escaped)
          return (values (coerce string 'string) :string)
        else 
          ;; just a random character, so collect it
          collect char into string))

(defmethod read-next-symbol ((self kimmo-reader) stream eof-value)
  ;; readtable-object's version of read-symbol.
  (loop for char = (peek-char-really stream eof-value)
        if (and (characterp (escape self))
                (characterp char)
                (char= char (escape self)))
          ;; found an escape char, so *NEXT* char definitely goes in the pname
          do (read-char stream (null eof-value) eof-value t)
          and collect (read-char stream (null eof-value) eof-value t) into print-name
        else if (or (eq char eof-value) ; either we get an eof
                    (and (characterp char) ; or char that ends symbols
                         (or (char= char (open-paren self))
                             (char= char (close-paren self))
                             (and (characterp (open-comment self)) ;***
                                  (char= char (open-comment self)))
                             (and (characterp (close-comment self))
                                  (char= char (close-comment self)))
                             (char= char (string-quote self))
                             (and (char= char #\Newline)
                                  (return-token self))
                             (member char (delimiter-chars self) :test #'char=)
                             (member char (whitespace-chars self) :test #'char=))))
          ;; this character breaks a symbol or is an EOF, so stop here
          return (let ((print-name (coerce print-name 'string)))
                   (if (every #'(lambda (char)
                                  (char-is-digit-p self char))
                              print-name)
                     ;; numbers are enabled, and these are all digits
                     (values (parse-integer print-name :radix (radix self))
                             :integer)
                     ;; just make a symbol
                     (values (intern print-name (package self))
                             :symbol)))
        else
          ;; just a random character, so collect it
          collect (read-char stream (null eof-value) eof-value t) into print-name))

(defmethod flush-next-comment ((self kimmo-reader) stream eof-value)
  ;; get rid of a comment
  (loop with level = 1
        ;; flush the open-comment.
        initially (read-char stream (null eof-value) eof-value t)
        for char = (read-char stream (null eof-value) eof-value t)
        if (and (characterp (escape self))
                (char= char (escape self)))
          ;; "I didn't mean it..."
          do (read-char stream (null eof-value) eof-value t)
        else if (char= char (open-comment self)) ;***
          ;; nested comment
          do (incf level)
        else if (char= char (close-comment self)) ;***
          ;; found end of comment
          do (decf level)
        while (> level 0)))

;;;
;;; Some higher-level read functions.
;;;

(defmethod read-n ((self kimmo-reader) n 
                   &optional
                   (stream *standard-input*)
                   (eof-value *eof-token*))
  ;; read n things, return a list of them.
  (loop for i below n
	while (listen stream)
        collecting (read-next-thing self stream eof-value)))

(defmethod read-until ((self kimmo-reader) endmarker
                       &optional
                       (stream *standard-input*)
                       (eof-value *eof-token*))
  (loop while (listen stream)
	for item = (read-next-thing self stream eof-value)
        when (eq item eof-value)
          do (error "End of file on ~S." stream)
        until (equal item endmarker)
        collecting item))

(defmethod repeat-read-until ((self kimmo-reader) endmarker
                              &optional 
                              (stream *standard-input*)
                              (eof-value *eof-token*))
  (loop for x = (read-until self endmarker stream eof-value)
        when x 
          return x))

(defmethod read-skipping-token ((self kimmo-reader) token
                                &optional
                                (stream *standard-input*)
                                (eof-value *eof-token*))
  (loop while (listen stream)
	as x = (read-next-thing self stream eof-value)
        unless (equal x token)
          return x))

;;;
;;; The reader that most of the rest of this program uses, and a test function.
;;;

(defvar *kimmo-reader* 
  (let ((reader (make-instance 'kimmo-reader :name "Kimmo Reader")))
    (disable-escapes reader)
    (enable-special-return reader '*CR*)
    reader))

(defun demo-kimmo-reader (&optional (stream *standard-input*))
  ;; type some things at it and see what it comes up with.  This should
  ;; be automated with a test string and with-input-from-string.
  (loop with reader = (make-instance 'kimmo-reader :name "Fred")
        doing (multiple-value-bind (token type)
                 (read-next-thing reader stream)
                (format *debug-io* "~&~S -> ~S" token type))))

(defun test-kimmo-reader ()
  ;; if it returns non-NIL, you're winning.  This is not an exhaustive test.
  ;; I didn't waste time trying to make it foolproof, since fools are so clever!
  (with-input-from-string (s "foo \"foo\" \"\\\"foo\\\"\" (a  b) Root")
    (let ((reader (make-instance 'kimmo-reader :name "Fred")))
      (and (eq (read-next-thing reader s) '|foo|)
           (string= (read-next-thing reader s) "foo")
           (string= (read-next-thing reader s) "\"foo\"") ;think about it...
           (equal (read-next-thing reader s) '(|a| |b|))
           ;; test the eof-terminates-symbols bug
           (eq (read-next-thing reader s) '|Root|)))))
